@@ -3,6 +3,7 @@
 #include <err.h>
 #include <errno.h>
 #include <inttypes.h>
+#include <locale.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -25,17 +26,17 @@ static char *name = "";
 #define pledge(p1,p2) 0
 #endif
 
-/* print `len' columns of characters. If string is shorter pad the rest
+/* format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
-void
-printutf8pad(FILE *fp, const char *s, size_t len, int pad)
+int
+utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
 	wchar_t w;
-	size_t col = 0, i, slen;
+	size_t col = 0, i, slen, siz = 0;
 	int rl, wc;
 
 	if (!len)
-		return;
+		return -1;
 
 	slen = strlen(s);
 	for (i = 0; i < slen && col < len + 1; i += rl) {
@@ -43,43 +44,74 @@ printutf8pad(FILE *fp, const char *s, size_t len, int pad)
 			break;
 		if ((wc = wcwidth(w)) == -1)
 			wc = 1;
-		col += (size_t)wc;
+		col += wc;
 		if (col >= len && s[i + rl]) {
-			fputs("\xe2\x80\xa6", fp);
-			break;
+			if (siz + 4 >= bufsiz)
+				return -1;
+			memcpy(&buf[siz], "\xe2\x80\xa6", 4);
+			return 0;
 		}
-		fwrite(&s[i], 1, rl, fp);
+		if (siz + rl + 1 >= bufsiz)
+			return -1;
+		memcpy(&buf[siz], &s[i], rl);
+		siz += rl;
+		buf[siz] = '\0';
 	}
-	for (; col < len; col++)
-		putc(pad, fp);
+
+	len -= col;
+	if (siz + len + 1 >= bufsiz)
+		return -1;
+	memset(&buf[siz], pad, len);
+	siz += len;
+	buf[siz] = '\0';
+
+	return 0;
 }
 
+/* Escape characters in text in geomyidae .gph format,
+   newlines are ignored */
 void
-trim(char *buf, size_t bufsiz, const char *src)
+gphtext(FILE *fp, const char *s, size_t len)
 {
-	size_t d = 0, i, len, s;
+	size_t i;
 
-	len = strlen(src);
-	for (s = 0; s < len && d < bufsiz - 1; s++) {
-		switch (src[s]) {
-		case '\t':
-			if (d + 8 >= bufsiz - 1)
-				goto end;
-			for (i = 0; i < 8; i++)
-				buf[d++] = ' ';
+	for (i = 0; *s && i < len; s++, i++) {
+		switch (*s) {
+		case '\r': /* ignore CR */
+		case '\n': /* ignore LF */
 			break;
-		case '|':
-		case '\n':
-		case '\r':
-			buf[d++] = ' ';
+		case '\t':
+			fputs("        ", fp);
 			break;
 		default:
-			buf[d++] = src[s];
+			fputc(*s, fp);
+			break;
+		}
+	}
+}
+
+/* Escape characters in links in geomyidae .gph format */
+void
+gphlink(FILE *fp, const char *s, size_t len)
+{
+	size_t i;
+
+	for (i = 0; *s && i < len; s++, i++) {
+		switch (*s) {
+		case '\r': /* ignore CR */
+		case '\n': /* ignore LF */
+			break;
+		case '\t':
+			fputs("        ", fp);
+			break;
+		case '|': /* escape separators */
+			fputs("\\|", fp);
+			break;
+		default:
+			fputc(*s, fp);
 			break;
 		}
 	}
-end:
-	buf[d] = '\0';
 }
 
 void
@@ -111,16 +143,15 @@ printtimeshort(FILE *fp, const git_time *intime)
 void
 writeheader(FILE *fp)
 {
-	char buf[256];
-
-	trim(buf, sizeof(buf), description);
-	if (buf[0] == 't' || buf[0] == '[')
-		fputc('t', fp);
-	fprintf(fp, "%s\n\n", buf);
+	if (description[0]) {
+		putchar('t');
+		gphtext(fp, description, strlen(description));
+		fputs("\n\n", fp);
+	}
 
 	fprintf(fp, "%-20.20s  ", "Name");
 	fprintf(fp, "%-50.50s  ", "Description");
-	fprintf(fp, "%-16.16s\n", "Last commit");
+	fprintf(fp, "%s\n", "Last commit");
 }
 
 int
@@ -155,16 +186,15 @@ writelog(FILE *fp)
 			*p = '\0';
 
 	fputs("[1|", fp);
-	trim(buf, sizeof(buf), stripped_name);
-	printutf8pad(fp, buf, 20, ' ');
+	utf8pad(buf, sizeof(buf), stripped_name, 20, ' ');
+	gphlink(fp, buf, strlen(buf));
 	fputs("  ", fp);
-	trim(buf, sizeof(buf), description);
-	printutf8pad(fp, buf, 50, ' ');
+	utf8pad(buf, sizeof(buf), description, 50, ' ');
+	gphlink(fp, buf, strlen(buf));
 	fputs("  ", fp);
 	if (author)
 		printtimeshort(fp, &(author->when));
-	trim(buf, sizeof(buf), stripped_name);
-	fprintf(fp, "|%s/%s/log.gph|server|port]\n", relpath, buf);
+	fprintf(fp, "|%s/%s/log.gph|server|port]\n", relpath, stripped_name);
 
 	git_commit_free(commit);
 err:
@@ -193,6 +223,8 @@ main(int argc, char *argv[])
 	if (pledge("stdio rpath", NULL) == -1)
 		err(1, "pledge");
 
+	setlocale(LC_CTYPE, "");
+
 	git_libgit2_init();
 
 	for (i = 1; i < argc; i++) {
@@ -232,7 +264,9 @@ main(int argc, char *argv[])
 		}
 		description[0] = '\0';
 		if (fp) {
-			if (!fgets(description, sizeof(description), fp))
+			if (fgets(description, sizeof(description), fp))
+				description[strcspn(description, "\t\r\n")] = '\0';
+			else
 				description[0] = '\0';
 			fclose(fp);
 		}
